/*
 * leds
 * Linux Editor for D 
 * S as in simple(superfluous) or supperb
 * 
 * by Antonio Monteiro
 */

module leds.CodeView;

private import dui.All;

private import c.stdio;
private import string;
private import path;

alias path filePath;

private import file;
private import syserror;
private import outbuffer;

private import leds.Workspace;
private import leds.FileView;
private import leds.Scintilla;
private import leds.FindAndReplace;

/**
 * A scrolled window that contains a scintilla widget
 */
public:
class CodeView : FileView , ScintillaListener
{
	
	protected:

	Scintilla scintilla;

	char[] previousFindString;

	public:
	
	/**
	 * creates a new empty CodeView
	 */
	this(Workspace workspace)
	{
		this(workspace, null);
	}

	/**
	 * creates a new CodeView and reads in a file
	 * @param fileName the fileName
	 */
	this(Workspace workspace, char[] fileName)
	{
		super(workspace, fileName);

		previousFindString = "";

		scintilla = new Scintilla();
		//scintilla.setRequest(-1,750);
		scintilla.send(SCI_STYLECLEARALL, 0, (sptr_t)0);
		scintilla.setDLex();

		// this will emite a message to the console:
		// cannot add non scrollable widget use gtk_scrolled_window_add_with_viewport() instead
		// we need to ignore that
		add(scintilla);
		setupNotifications();
		if ( fileName !== null )
		{
			loadFile(fileName);
		}
	}

	/**
	 * loads a file into the view
	 * @param fileName the file to load
	 * @return true if sucesseful
	 */
	bit loadFile(char[] fileName)
	{
		this.fileName = fileName.dup;
		char[] input;
		try
		{
			input = cast(char[])file.read(fileName);
			scintilla.send(SCI_STYLECLEARALL, 0, (sptr_t)0);
			scintilla.setDLex();
			scintilla.send(SCI_INSERTTEXT, 0, (sptr_t)(char*) input);
			scintilla.send(SCI_EMPTYUNDOBUFFER,0 ,0 );
		}
		catch ( Error e)
		{
			e.print();
		}
		return true;
	}

	/**
	 * Indicates if the text was modified
	 * @return true if the text was modified
	 */
	bit getModified()
	{
		return false;
	}
	
	/**
	 * indicates if this buffer is a empty never used buffer
	 * @return true if this instance should be loaded with a new file
	 */
	bit getEmpty()
	{
		return !getModified() && (fileName === null);
	}
	 
	/**
	 * checks if the buffer needs to be saved
	 * @return true if the buffer can be closed
	 */
	bit allowClose()
	{
		bit allow = true;
		printf("allow\n");
		if ( getModified() )
		{
			MessageDialog d = new MessageDialog(workspace, 0,
											DialogType.QUESTION,
											DialogButtons.NONE ,
											"File Not save!\nSave Now?");
			d.addButton(Stock.NO,Response.NO);
			d.addButton(Stock.YES,Response.YES);
			d.addButton(Stock.CANCEL,Response.CANCEL);
			int responce = d.run();
			if ( responce == Response.YES )
			{
				saveFile();
			}
			else if ( responce == Response.CANCEL )
			{
				allow = false;
			}
			d.destroy();
		}
		return allow;
	}
	
	/**
	 * saves the current file into the hard drive
	 * @return true if sucesseful
	 */
	bit saveFile()
	{
printf("CodeView.saveFile 1\n");
		OutBuffer buf = new OutBuffer();
printf("CodeView.saveFile 2\n");

		int textLength = scintilla.send(SCI_GETLENGTH,0,0)+1;
printf("CodeView.saveFile 3\n");
		
		//printf("file size = %d\n",textLength);
		
		char[] text = new char[textLength];
printf("CodeView.saveFile 4\n");
		char* p = &text[0];	
printf("CodeView.saveFile 5\n");
		scintilla.send(SCI_GETTEXT,textLength,(sptr_t)p);
printf("CodeView.saveFile 6\n");
		//printf("p = \n%s\n",p);
		//printf("text = \n%.*s\n",text);
		text.length = text.length-1;
printf("CodeView.saveFile 7\n");
		
		buf.write(text);
printf("CodeView.saveFile 8\n");
		try
		{
printf("CodeView.saveFile 9\n");
			file.write(fileName,(byte[])buf.toBytes());
printf("CodeView.saveFile 9a\n");
		}
		catch ( FileError fe )
		{
printf("CodeView.saveFile 10\n");
			printf("CodeView.saveFile FileError\n");
			workspace.saveAs(this, "Error! Please entre file name");
printf("CodeView.saveFile 10a\n");
		}
		catch ( SysError se )
		{
printf("CodeView.saveFile 11\n");
			printf("CodeView.saveFile SysError\n");
			workspace.saveAs(this, "Error! Please entre file name");
printf("CodeView.saveFile 11a\n");
		}
		catch ( Error e )
		{
printf("CodeView.saveFile 12\n");
			printf("CodeView.saveFile Error\n");
			workspace.saveAs(this, "Error! Please entre file name");
printf("CodeView.saveFile 12a\n");
		}

		return true;
	}
	
	/**
	 * renames and saves the current file into the hard drive
	 * @return true if sucesseful
	 */
	bit saveFile(char[] fileName)
	{
printf("CodeView.saveFile a1\n");
		this.fileName = fileName.dup;
printf("CodeView.saveFile a2\n");
		return saveFile();
	}
	
	/**
	 * renames our current file name
	 */
	void rename(char[] fileName)
	{
	}

	/**
	 * execute an action if action is known.
	 * @param action
	 * @return true if the action was processed (even if the execution failled)
	 */
	public bit execAction(char[] action)
	{
		bit processed = true;
		switch(action)
		{
			case "view.zoom.in":	scintilla.send(SCI_ZOOMIN, 0, 0); break;
			case "view.zoom.out":   scintilla.send(SCI_ZOOMOUT, 0, 0); break;
			
			case "toolbar.undo":	scintilla.send(SCI_UNDO, 0, 0); break;
			case "toolbar.redo":	scintilla.send(SCI_REDO, 0, 0); break;

			case "toolbar.copy":	scintilla.send(SCI_COPY, 0, 0); break;
			case "toolbar.cut":	 	scintilla.send(SCI_CUT, 0, 0); break;
			case "toolbar.paste":	scintilla.send(SCI_PASTE, 0, 0); break;

			case "toolbar.findText": findText(); break;

			case "toolbar.gotoLine": 
				workspace.getGotoLine().show(scintilla.send(SCI_GETLINECOUNT, 0, 0));
				break;
				
			case "lexer.D": scintilla.setDLex();	break;
			case "lexer.makefile": scintilla.setMakefileLex();	break;
				
			default:
				processed = false;
			break;
		}
		return processed;
	}

	/**
	 * find for the text on the toolbar find text entry
	 */
	void findText()
	{
		findText(workspace.getFindText());
	}

	/**
	 * find for a temporary text string
	 */
	void findText(char[] findStr)
	{
		if ( findStr === null || findStr == "" )
		{
			findStr = previousFindString;
		}
		if ( (findStr!==null) && (findStr.length>0) )
		{
			findReplace(findStr, null, false, true, 0);
		}
	}
	
	/**
	 * goto a line
	 */
	void gotoLine(int lineNumber)
	{
		if ( lineNumber > 0 )
		{
			scintilla.send(SCI_GOTOLINE, lineNumber-1, 0);
		}
	}
	
	/**
	 * Finds or replaces text
	 * @param findStr the string to find
	 * @param replaceStr the string to replace, the new string
	 * @param replace if true the replaceStr should replace the selected text
	 * @param find if true after replace do another find
	 * @param flags the find mode @see Scintilla
	 */
	void findReplace(char[] findStr, char[] replaceStr, bit replace, bit find,int flags)
	{
		
		if ( replace )
		{
			
		}
		
		if ( find )
		{
			previousFindString = findStr.dup;
			scintilla.find(findStr, flags);
		}
		
	}

	void setupNotifications()
	{
		scintilla.addScintillaListener((ScintillaListener)this);
		scintilla.send(SCN_KEY,0 ,0);
	}
	
	/**
	 * Processes sintilla notifications
	 */
	void scintillaCallback(SCNotification* scNotification)
	{
		switch(scNotification.nmhdr.code)
		{
			case SCN_KEY:
				if ( scNotification.modifiers == 2 )
				{
					switch ( scNotification.ch )
					{
						case 'S': workspace.saveCurrent();  break;
						case 'W': workspace.closeView();	break;
						case 'K': findText(scintilla.getSelected()); break;
						
						default:
							printf("key %c %X \n",scNotification.ch,scNotification.modifiers);
							break;
					}
				}
				break;

			default:
				//printf("CodeView.scintillaCallback\n");
				break;
		}
		int pos = scintilla.send(SCI_GETCURRENTPOS,0,0);
		workspace.getStatusbar().setLinesCol(
			scintilla.send(SCI_LINEFROMPOSITION,pos , 0)+1,
			scintilla.send(SCI_GETLINECOUNT, 0, 0),
			scintilla.send(SCI_GETCOLUMN, pos, 0)
		);
		
	}

}