/*
 * leds
 * Linux Editor for D 
 * S as in simple(superfluous) or supperb
 * 
 * by Antonio Monteiro
 */

module leds.Project;

private import dui.All;

private import leds.Leds;
private import leds.Workspace;
private import leds.Properties;

private import path;
private import string;

const char [] projectHeaderMessage = 
"#"
"\n# leds project file"
"\n#"
"\n"
"\n#"
"\n# This file was generated automatically by leds"
"\n# editing this file is not recomended"
"\n#"
"\n"
"\n"
;


public:

enum ProjectType
{
	GENERIC,
	CONSOLE,
	DUI,
	LIB
};

/**
 * This defines a project.
 * A project has:<br>
 * - name<br>
 * - base directory<br>
 * - dependencies (libraries or other projects)
 * - target type<br>
 * - target names (execs or libs, for instance)<br>
 * - build (for instance a Makefile or a digc (digc not implemented))<br>
 * A project will be able to:<br>
 * - save it's state<br>
 * - load it's state<br>
 * - build the targets<br>
 * - allow the user to define it's parameters<br>
 * - create a Makefile from it's parameters<br>
 *<br>
 * project wil be save on disk at the base directory workspace.getDataDir()/projects
 *\todo extract the project interface creates an Abstract and subclass fro each type of project
 */
public class Project : PropertiesGUICaller
{
	
	private:

	bit validProject = true; // volatille
	
	Workspace workspace;
	ProjectType type;

	Properties props;

	bit addAfterEdit = false;
	
	public:

	/**
	 * Creates a new project.
	 * the project will try to load it's definitions.
	 * If no definition is found the interactive mode
	 * will be entered.
	 * @param workspace
	 * @param name the project name
	 */
	this(Workspace workspace, char[] name)
	{
		setWorkspace(workspace);
		setProps();
		setName(name);
		if ( !load() )
		{
			setActive(true);
			guiSetup();
		}
	}
	
	/**
	 * Creates a new project.
	 * the project will try to load it's definitions.
	 */
	this(Workspace workspace, char[] name, char[] type)
	{
		setWorkspace(workspace);
		setProps();
		setName(name);
		setType(type);
		guiSetup();
	}

	/**
	 * creates a project and tries to load
	 * @param filename the project definition file
	 */
	this(char[] filename)
	{
		setProps();
		setActive(false);
		props.load(filename);
	}
	
	void setName(char[] name)
	{
		props.set(0,name);
	}
	char[] getName()
	{
		return props.get(0,"New project");
	}
	void setType(char[] type)
	{
		props.set(1,type);
	}
	char[] getType()
	{
		return props.get(1,"DUI");
	}
	char[] getBaseDir()
	{
		return props.get(3,Leds.leds.getUserHome());
	}

	char[] getPropValue(char[] key, char[] defaultValue)
	{
		return props.get(key,defaultValue);
	}
		
	void setProps()
	{
		props = new Properties();
		props.add(PropsTypes.TEXT, "name:", "", "Name","","");
		props.add(PropsTypes.TEXT, "type:", "DUI", "Type","","");
		props.add(PropsTypes.YES_NO, "active:", "", "Active","","");
		props.add(PropsTypes.DIRECTORY, "baseDir:", "", "Base dir","","");
		props.add(PropsTypes.YES_NO, "showHidden:", "no", "Show hidden files","","");
		props.add(PropsTypes.TEXT, "runCommand:", "", "Run command","","");
		
		props.add(PropsTypes.TEXT, "makeCommand:", "make", "make","","Build");
		props.add(PropsTypes.FILE, "dmdCommand:", "dmd", "DMD location","","Build");
		props.add(PropsTypes.MULTY_DIRECTORY, "importDirs:", "", "Imports","","Build");
		props.add(PropsTypes.MULTY_TEXT, "dependencies:", "", "Dependencies","","Build");
		props.add(PropsTypes.MULTY_TEXT, "targets:", "", "Targets","","Build");
		
		props.add(PropsTypes.TEXT, "userDocCommand:", "", "User Documentation tool","","Doc");
		props.add(PropsTypes.TEXT, "develDocCommand:", "", "Developer Documentation tool","","Doc");
		props.add(PropsTypes.TEXT, "privateDocCommand:", "", "Private Documentation tool","","Doc");
		
		props.add(PropsTypes.TEXT, "backupCommand:", "", "backup tool","","repository");
	}
	
	void setWorkspace(Workspace workspace)
	{
		this.workspace = workspace;
	}
	/**
	 * inquires the active state of this project
	 * @return the active state of this project
	 */
	bit isActive()
	{
		return props.get(2,"yes") == "yes";
	}
	
	/**
	 * Sets the active flags of this project
	 * @param active the new active state
	 */
	void setActive(bit active)
	{
		props.set(2, !active ? "no" : "yes");
	}

	bit isValid()
	{
		return validProject;
	}

	bit getShowHiddenFiles()
	{
		return props.get("showHiddenFiles:","no") == "yes";
	}
	public bit save()
	{
		return props.save(Leds.leds().getProjectFileName(getName()),projectHeaderMessage);
	}
	
	public bit load()
	{
		return props.load(Leds.leds().getProjectFileName(getName()));
	}
	
	/**
	 * executes the build command for this project
	 */
	public void build()
	{
	}
	
	public void guiSetup()
	{
		guiSetup(true);
	}

	Properties getProps()
	{
		return props;
	}
	
	void propertiesEditDone(Properties newProperties)
	{
		props = newProperties;
		save();
		if ( workspace != null && addAfterEdit )
		{
			workspace.addProject(this);
		}
	}

	public void guiSetup(bit addAfterEdit)
	{
		this.addAfterEdit = addAfterEdit;
		new PropertiesGUISetup(this, workspace);
	}
	
	public void createMakefile()
	{
	}

	public void dump()
	{
		props.dump();
	}
		
};
