/*
 * leds
 * Linux Editor for D 
 * S as in simple(superfluous) or supperb
 * 
 * by Antonio Monteiro
 */

module leds.Properties;

private import dui.All;

private import leds.Workspace;

private import file;
private import outbuffer;
private import string;

enum PropsTypes
{
	TEXT,
	YES_NO,
	INT,
	BIT,
	FILE,
	DIRECTORY,
	ICON,
	MULTY_TEXT,
	MULTY_FILE,
	MULTY_DIRECTORY,
}

/*
 * Represnts one propertie
 */
class Props
{
	char[] label;
	char[] name;
	char[] value;
	/** a regexp to match the value entered */
	char[] validator;
	/** the prop tab. if "" or null a "Basic" tab will be used */
	char[] tabName;
	
	PropsTypes type;
	
	this(PropsTypes type, char[] name, char[] value, char[] label )
	{
		this(type, name, value, label, "", "");
	}
	this(PropsTypes type, char[] name, char[] value, char[] label, char[] validator, char[] tabName)
	{
		this.label = label;
		this.name = name;
		this.value = value;
		this.type = type;
		this.validator = validator;
		this.tabName = tabName;
	}

	this(Props prop)
	{
		label = prop.label.dup;
		name = prop.name.dup;
		value = prop.value.dup;
		type = prop.type;
	}
}

/**
 * Points to a prop in error
 */
class PropsError
{
	int field;
	char[] message;
	this(int field, char[] message)
	{
		this.field = field;
		this.message = message;
	}
}

/**
 * A set os properties that can be read and write to a file
 */
class Properties
{

	/** properties by name */
	int order = 0;
	Props[char[]] props;
	/** order is imprtant on the text definition file */
	char[][int] propsOrder;
	
	PropsError[] errors;

	/**
	 * Creates a new properties set
	 */
	this()
	{
	}
	/**
	 * Copy ctor
	 */
	this(Properties properties)
	{
		for ( int i = 0 ; i<properties.length ; i++)
		{
			Props prop = properties.getProp(i);
			add(prop.type, prop.name, prop.value, prop.label , prop.validator , prop.tabName);
		}
	}
	
	int length()
	{
		return props.length;
	}
	
	void add(PropsTypes type, char[] name, char[] value, char[] label, char[] validator, char[] tabName)
	{
		props[name] = new Props(type, name, value, label, validator, tabName );
		propsOrder[order++] = name;
	}
	

	Props getProp(int index)
	{
		if ( index in propsOrder )
		{
			return props[propsOrder[index]];
		}
		else
		{
			return null;
		}
	}
	
	void set(int index, char[] value)
	{
		if ( index in propsOrder )
		{
			set(propsOrder[index],value);
		}
	}
	
	char[] get(int index, char[] defaultValue)
	{
		if ( index in propsOrder )
		{
			return get(propsOrder[index],defaultValue);
		}
	}

	void set(char[] key, char[] value)
	{
		if ( key in props )
		{
			props[key].value = value;
		}
	}
	
	char[] get(char[] key, char[] defaultValue)
	{
		if ( key in props )
		{
			return props[key].value;
		}
		else
		{
			return defaultValue;
		}
	}

	bit validate()
	{
		errors.length = 0;

		if ( get(0,"").length==0 )
		{
			errors ~= new PropsError(0,"You must enter a project name.");
		}
		
		if ( get(3,"").length==0 )
		{
			errors ~= new PropsError(3,"You must enter the project base directory.");
		}
		
		return errors.length == 0;
	}


	/**
	 * dumps all this project definitions into a file.
	 * All existing definition on the file will be ignored
	 */
	public bit save(char[] filename, char[] headerMessage)
	{

		OutBuffer buf = new OutBuffer();

		buf.write(headerMessage);

		for ( int i = 0 ; i<length ; i++)
		{
			Props prop = getProp(i);
			buf.write(prop.name~"\t"~prop.value~"\n");	
		}
		
		file.write(filename,(byte[])buf.toBytes());
		return true;
	}

	/**
	 * loads this project definitions from a file
	 */
	public bit load(char[] filename)
	{
		printf("ProjectProperties.load() %.*s\n",filename);
		if ( props.length < 1 ) return true;
		
		bit inError = false;

		char[] input;
		try
		{
			input = cast(char[])file.read(filename);
		}
		catch (FileError error)
		{
			return false;
		}
		char[][]tokens;
		
		int lineNumber = 0;

		char[] getToEOL(int pos, char[] line)
		{
			if ( line.length>pos )
			{
				return line[pos..line.length];
			}
			return "";
		}

		bool getYesNo(char[][] tokens)
		{
			bit b;
			if ( tokens.length>1 )
			{
				return tokens[1] == "yes";
			}
			else
			{
				b = false;
				return b;
			}
		}
		
		void processToken(char[][] tokens, char[] line)
		{
			if ( tokens[0] in props )
			{
				Props prop = props[tokens[0]];
				switch ( prop.type )
				{
					case PropsTypes.MULTY_TEXT:
					case PropsTypes.MULTY_FILE:
					case PropsTypes.MULTY_DIRECTORY:
						break;

					case PropsTypes.YES_NO:
						if ( tokens.length > 1 )
						{
							prop.value = tokens[1];
						}
						break;
						
					default:
						prop.value = getToEOL(prop.name.length+1,line); break;
						break;
				}
			}
		}

	printf("Properties.load before line by line\n");		
		foreach(char[] line; string.splitlines(input))
		{
			++lineNumber;
	printf("properties.load line = %.*s\n",line);
			tokens = string.split(line);
			if ( tokens.length> 0 && tokens[0][0] != '#' )
			{
				processToken(tokens,line);
			}
		}

		dump();
		
		return true;
		
	}

	void dump()
	{
		foreach(Props prop; props)
		{
			printf("%.*s\t%.*s\n",prop.name,prop.value);
		}
	}
	
	
		
};

/**
 * The caller for the properties
 */
public:
interface PropertiesGUICaller
{
	Properties getProps();
	void propertiesEditDone(Properties newProperties);
};

/**
 * User interface for the properties
 */
public:
class PropertiesGUISetup : MainWindow , MenuItemListener , ButtonClickedListener , WindowListener
{

	PropertiesGUICaller caller;	
	Properties originalProps;
	Properties editProps;
	bit addAfterSetup = true;

	struct TabTable
	{
		Table table;
		int row;
	}
	

	TabTable*[char[]] tabTables;
	
	Widget[] entries;
	
	this(PropertiesGUICaller caller, Workspace workspace)
	{
		this.caller = caller;
		this.originalProps = new Properties(caller.getProps());
		editProps = new Properties(originalProps);
		this.addAfterSetup = addAfterSetup;
		super("Project Definition");
		setup();
		setTransientFor(workspace);
		show();
	}

	~this()
	{
		destroy();
	}

	/**
	 * sets the user interface values from the project values
	 */
	void setEntryFromProps(Properties props)
	{
		printf("setEntryFromProject 1 \n");
		for ( int i = 0; i<entries.length ; i++)
		{
			Props prop = props.getProp(i);
		printf("setEntryFromProject 2 %.*s\n",prop.name);
			switch( prop.type )
			{
				case PropsTypes.MULTY_TEXT:
		printf("setEntryFromProject 3\n");
					((Entry)entries[i]).setText(prop.value);
					break;

				case PropsTypes.MULTY_FILE:
		printf("setEntryFromProject 4\n");
					((Entry)entries[i]).setText(prop.value);
					break;

				case PropsTypes.MULTY_DIRECTORY:
					((TextView)entries[i]).getBuffer.setText(prop.value);
					break;

				case PropsTypes.YES_NO:
		printf("setEntryFromProject 5\n");
					((CheckButton)entries[i]).setActive(prop.value == "yes");
					break;
					
				default:
		printf("setEntryFromProject 6\n");
					((Entry)entries[i]).setText(prop.value);
					break;
			}
		}
		printf("setEntryFromProject exit\n");
		
	}

	/**
	 * sets the project values from the user interface values
	 */
	void setPropsFromEntry(Properties props)
	{

		printf("starting set project from entry\n");
		for ( int i = 0; i<entries.length ; i++)
		{
			Props prop = props.getProp(i);
			switch( prop.type )
			{
				case PropsTypes.MULTY_TEXT:
					prop.value = ((Entry)entries[i]).getText().dup;
					break;

				case PropsTypes.MULTY_FILE:
					prop.value = ((Entry)entries[i]).getText().dup;
					break;

				case PropsTypes.MULTY_DIRECTORY:
					prop.value = ((TextView)entries[i]).getBuffer().getText().dup;
					break;

				case PropsTypes.YES_NO:
					prop.value = !((CheckButton)entries[i]).getActive ? "no" : "yes";
					break;
					
				default:
					prop.value = ((Entry)entries[i]).getText().dup;
					break;
			}
		}
		printf("Just finished set project from entry\n");
		props.dump();
	}

	private TabTable* getTabTable(char[] tabName, Notebook notebook)
	{
		TabTable* tabTable;
		if ( tabName == null || tabName == "" )
		{
			tabName = "Basic";
		}
		if ( tabName in tabTables )
		{
			tabTable = tabTables[tabName];
		}
		else
		{
			tabTable = new TabTable;
			tabTable.table = new Table(3,1,false);
			tabTable.row = 0;
			tabTables[tabName] = tabTable;
			notebook.appendPage(tabTable.table,tabName);
		}
		return tabTable;
	}
	
	private void setup()
	{
		Button button;
		Dispatcher disp = Dispatcher.getDispatcher();

		OptionMenu projectType = new OptionMenu();
		Menu menu = new Menu();
		menu.append(new MenuItem(this,"Generic","projectType.generic"));
		menu.append(new MenuItem(this,"Console","projectType.Console"));
		menu.append(new MenuItem(this,"DUI","projectType.DUI"));
		menu.append(new MenuItem(this,"Lib","projectType.lib"));

		projectType.setMenu(menu);
		projectType.setHistory(2);

		Notebook notebook = new Notebook();
		
		VBox vbox = new VBox(false,4);
		
		vbox.packStart(new Label(editProps.get("name:","")),false,false,2);

		int row = 0;

		void addTable(TabTable* tabTable, char[] lbl, Widget entry, char[] btn, char[] act)
		{
			tabTable.table.attach(
					new Label(lbl), 0, 1, 
					tabTable.row, tabTable.row+1, 
					AttachOptions.SHRINK,AttachOptions.SHRINK,2,2);
			tabTable.table.attach(entry,1,2,
					tabTable.row,tabTable.row+1,
					AttachOptions.FILL,AttachOptions.FILL,2,2);
			if ( btn !== null )
			{
				Button button = new Button(btn);
				tabTable.table.attach(
						button,2,3,
						tabTable.row,tabTable.row+1,
						AttachOptions.SHRINK,AttachOptions.SHRINK,2,2);
				Dispatcher.getDispatcher().addButtonClickedListener(this,button,act);
			}
			tabTable.row++;
		}

		TabTable* tabTable;

		for ( int i = 0; i<editProps.length ; i++)
		{
			Props prop = editProps.getProp(i);
			printf("props %d = %.*s\n",i,prop.name);
			tabTable = getTabTable(prop.tabName, notebook);
			switch ( prop.type )
			{
				case PropsTypes.MULTY_TEXT:
					Entry entry = new Entry(prop.value);
					addTable(tabTable,prop.label,entry,null,null); 
					entries ~= entry;
					break;

				case PropsTypes.MULTY_FILE:
					Entry entry = new Entry(prop.value);
					addTable(tabTable,prop.label,entry,null,null); 
					entries ~= entry;
					break;

				case PropsTypes.MULTY_DIRECTORY:
					TextView tv = new TextView();
					addTable(tabTable,prop.label,tv,null,null); 
					entries ~= tv;
					break;

				case PropsTypes.YES_NO:
					CheckButton cb = new CheckButton();
					cb.setActive(prop.value == "yes");
					addTable(tabTable,prop.label,cb,null,null); 
					entries ~= cb;
					break;
					
				default:
					Entry entry = new Entry(prop.value);
					addTable(tabTable,prop.label,entry,null,null); 
					entries ~= entry;
					break;
			}
		}

		vbox.packStart(notebook,false,false,2);

		// butons
		
		HButtonBox buttonBox = HButtonBox.createActionBox();
		
		button = new Button(Stock.OK);
		disp.addButtonClickedListener(this,button,"action.ok");
		buttonBox.packStartDefaults(button);
		button = new Button(Stock.CANCEL);
		disp.addButtonClickedListener(this,button,"action.cancel");
		buttonBox.packStartDefaults(button);
		
		vbox.packStart(new HSeparator(),false,false,0);
		vbox.packStart(buttonBox,false,false,2);
		
		add(vbox);
		setEntryFromProps(editProps);
	}

	private void editError(Properties props)
	{
		int maxMessages = 1;
		int countMessages = 0;
		foreach(PropsError error; props.errors)
		{
			if ( countMessages < maxMessages )
			{
				MessageDialog d = new MessageDialog(this, 0,
										DialogType.ERROR,DialogButtons.OK,
										error.message
										);
				++countMessages;
				d.run();
				d.destroy();
				entries[error.field].grabFocus();
			}
			
		}
	}
	
	void buttonClickedCallback(Button button, char[] action)
	{
		bit close = false;
		switch ( action )
		{
			case "selectBaseDir":
				char [] dir = selectDir();
				//baseDirEntry.setText(dir);
				break;

			case "action.ok":
				setPropsFromEntry(editProps);
				if ( editProps.validate() )
				{
		printf("Helloooooooooooooo 1\n");
					caller.propertiesEditDone(editProps);
		printf("Helloooooooooooooo 1\n");
					close = true;
				}
				else
				{
					editError(editProps);
				}
				break;
				
			case "action.cancel":
				close = true;
				break;
				
			default:
				break;
		}
		if ( close )
		{
			destroy();
		}
	}

	private FileSelection fileSelection;
	
	char[] selectDir()
	{
		if ( fileSelection === null )
		{
			fileSelection = new FileSelection("Select Directory");
		}
		fileSelection.setMultiple(false);
		fileSelection.run();
		fileSelection.hide();
		return fileSelection.getFileName();
	}
	
	void activateItemCallback(MenuItem menuItem, char [] action)
	{
		printf("activateItemCallback %.*s\n",action);
	}
	
	void activateCallback(MenuItem menuItem, char [] action)
	{
		printf("activateCallback %.*s\n",action);
	}

	
	bit windowDeleteCallback(Window window, Event event)
	{
		return true;
	}
	bit windowDestroyCallback(Window window, Event event)
	{
		return false;
	}

}