/* leds
 * Linux Editor for D 
 * S as in simple(superfluous) or supperb
 * 
 * by Antonio Monteiro
 */

module leds.Workspace;

private:
private import dui.All;

private import path;
alias path filePath;
private import string;

private import leds.Leds;
private import leds.FileView;
private import leds.CodeView;
private import leds.ImageView;
private import leds.Project;
private import leds.SystemConsole;
private import leds.Docker;
private import leds.ProjectView;
private import leds.StatusBar;
private import leds.GotoLineDialog;
private import leds.FindAndReplace;

private import ext.dconf.DConf;

/**
 * A main window for leds
 */
public:
class Workspace : MainWindow , MenuItemListener , ButtonClickedListener
{

	Leds parentLeds;
	FileView[] fileViews;
	VBox mainBox;

	Docker dockLeft;	
	Docker dockTop;	
	Docker dockRight;	
	Docker dockBottom;	
	Docker dockCode;

	HPaned leftMainPane;
	VPaned bottomMainPane;

	ProjectView projectView;
	SystemConsole systemConsole;
	StatusBar statusbar;

	// we want our dialogs to be presistent
	FindAndReplace findAndReplace;
	GotoLineDialog gotoLineDialog;

	Combo findTextCombo;

	/**
	 * the file selection dialog.
	 * reusing the same dialog will remember the last directory used
	 */
	FileSelection fileSelection;
	
	DConf dConf;
	
	public:

	/**
	 * ctor. must more activate the constructor name change loby
	 */
	this(Leds parentLeds, bit openProjects)
	{
		super("leds - Linux Editor for D");
		this.parentLeds = parentLeds;
		dConf = new DConf();

		printf("Workspace this 1\n");
		setup();
		printf("Workspace this 2\n");
		init();
		printf("Workspace this 3\n");
		if ( openProjects )
		{
			projectView.openActiveProjects();
		}
		printf("Workspace this 4\n");

	}
	
	/**
	 * Creates the window components and opens the command line files
	 * @param filesToOpen the files to open
	 */
	void setup()
	{
		setBorderWidth(0);

		mainBox = new VBox(false,0);
		add(mainBox);

		mainBox.packStart(getMenubar(this),false,false,0);		
		mainBox.packStart(getToolbar(this),false,false,0);		
		mainBox.packStart(getViews(),true,true,0);		
		mainBox.packStart(getStatusbar(),false,false,0);		

		setDefaultDockers();
		
	}
	
	/**
	 * reads the preferences parameters and aplies them
	 */
	void init()
	{
		int width = dConf.getInt("/apps/leds/init/main/width");
		int height = dConf.getInt("/apps/leds/init/main/heigth");
		int top = dConf.getInt("/apps/leds/init/main/top");
		int left = dConf.getInt("/apps/leds/init/main/left");
		int posLeft = dConf.getInt("/apps/leds/init/main/lPos");
		int posBottom = dConf.getInt("/apps/leds/init/main/bottomPos");

		if ( width == 0 || height == 0 )
		{
			width = -1;
			height = 550;
		}
		setDefaultSize(width,height);		
		move(left,top);
		
		leftMainPane.setPosition(posLeft);
		bottomMainPane.setPosition(posBottom);
		
	}

	/**
	 * saves the preferences and exits
	 * @param code the code that generated this exit request
	 * @param force if true the program will exit anyway
	 * @return true it the exit should proced false if the application should not exit
	 */
	protected bit exit(int code, bit force)
	{

		int width;
		int height;
		getSize(&width, &height);		

		dConf.setInt("/apps/leds/init/main/width",width);
		dConf.setInt("/apps/leds/init/main/heigth",height);
		int left;
		int top;
		getPosition(&left,&top);
		dConf.setInt("/apps/leds/init/main/left",left);
		dConf.setInt("/apps/leds/init/main/top",top);
		
		int pos = leftMainPane.getPosition();
		dConf.setInt("/apps/leds/init/main/lPos",pos);

		pos = bottomMainPane.getPosition();
		dConf.setInt("/apps/leds/init/main/bottomPos",pos);

		return true; //parentLeds(code, force);
	}
	
	/**
	 * creates the minimum workspace
	 */
	void setDefaultDockers()
	{
		dockCode.dock(new CodeView(this));
		dockLeft.dock(getProjectView());
		dockBottom.dock(getSystemConsole());
	}
	
	/**
	 * Creates the standard main menu
	 */
	public Widget getMenubar(MenuItemListener listener)
	{
		MenuBar menubar = new MenuBar();
		Menu menu;
		MenuItem item;
		Menu submenu;
		
		// must init dispatcher here or it want work on the MenuItem
		Dispatcher dispatcher = Dispatcher.getDispatcher();
		
		menu = menubar.append("_File");
		menu.append(new MenuItem(listener,"_New","file.new"));

		submenu = menu.appendSubmenu("New from _template");
		submenu.append(new MenuItem(listener, "_D","file.newFromTemplate.d"));

		menu.append(new MenuItem(listener,"New _project","file.newProject"));
		
		menu.append(new SeparatorMenuItem());
		menu.append(new MenuItem(listener,"_Open","file.open"));
		menu.append(new MenuItem(listener,"Open in _Area","file.openInArea"));
		menu.append(new MenuItem(listener,"Recent","file.recent"));
		menu.append(new MenuItem(listener,"Clear Recent","file.clearRecent"));
		menu.append(new SeparatorMenuItem());
		menu.append(new MenuItem(listener,"_Close","file.close"));
		menu.append(new MenuItem(listener,"Close All","file.closeAll"));
		menu.append(new SeparatorMenuItem());
		menu.append(new MenuItem(listener,"Save","file.save"));
		menu.append(new MenuItem(listener,"Save All","file.saveAll"));
		menu.append(new MenuItem(listener,"Save As","file.saveAs"));
		menu.append(new SeparatorMenuItem());
		menu.append(new MenuItem(listener,"E_xit","file.exit"));

		menu = menubar.append("_Edit");
		menu.append(new MenuItem(listener,"_Undo","edit.undo"));
		menu.append(new MenuItem(listener,"_Redo","edit.redo"));
		menu.append(new SeparatorMenuItem());
		menu.append(new MenuItem(listener,"Copy","edit.copy"));
		menu.append(new MenuItem(listener,"Append Copy","edit.appendCopy"));
		menu.append(new MenuItem(listener,"Cut","cut"));
		menu.append(new MenuItem(listener,"Append Cut","edit.appendCut"));
		menu.append(new MenuItem(listener,"Paste","edit.paste"));
		menu.append(new MenuItem(listener,"Filtered Paste","edit.filteredPaste"));
		menu.append(new MenuItem(listener,"Insert Paste","edit.insertPaste"));
		menu.append(new SeparatorMenuItem());
		menu.append(new MenuItem(listener,"Select _All","edit.selectedAll"));
		menu.append(new SeparatorMenuItem());
		menu.append(new MenuItem(listener,"Goto Line","edit.gotoLine"));
		menu.append(new MenuItem(listener,"Delete Line","edit.deleteLine"));
		menu.append(new MenuItem(listener,"Date & time","edit.DateTime"));
		menu.append(new SeparatorMenuItem());
		menu.append(new MenuItem(listener,"Preferences","edit.preferences"));
		//dispatcher.addMenuItemListener(listener,itemFind,"fedit.find");

		menu = menubar.append("_Search");
		menu.append(new MenuItem(listener,"_Find","search.find"));
		menu.append(new MenuItem(listener,"_Replace","search.replace"));
		menu.append(new MenuItem(listener,"Find _All","search.findAll"));
		menu.append(new MenuItem(listener,"Find next","search.findNext"));
		menu.append(new MenuItem(listener,"Replace next","search.replaceNext"));
		menu.append(new MenuItem(listener,"Replace selection","search.replaceSelection"));

		menu = menubar.append("Sour_ce");

		menu = menubar.append("_Projects");
		menu.append(new MenuItem(listener,"_New","projects.newProject"));
		menu.append(new MenuItem(listener,"_List","projects.list"));
		menu.append(new SeparatorMenuItem());
		menu.append(new MenuItem(listener,"_Build","project.build"));
		menu.append(new MenuItem(listener,"Build _All","project.buildAll"));
		menu.append(new SeparatorMenuItem());
		menu.append(new MenuItem(listener,"_Run","project.run"));

		menu = menubar.append("_View");
		submenu = menu.appendSubmenu("Toggle Panes");
		submenu.append(new MenuItem(listener,"Toggle _left panel","view.toggleLeftPanel"));
		submenu.append(new MenuItem(listener,"Toggle _top panel","view.toggleTopPanel"));
		submenu.append(new MenuItem(listener,"Toggle _right panel","view.toggleRightPanel"));
		submenu.append(new MenuItem(listener,"Toggle _bottom panel","view.toggleBottomPanel"));

		submenu = menu.appendSubmenu("Tabs position");
		submenu.append(new MenuItem(listener,"Left","view.tabsPosition.left"));
		submenu.append(new MenuItem(listener,"Top","view.tabsPosition.top"));
		submenu.append(new MenuItem(listener,"Right","view.tabsPosition.right"));
		submenu.append(new MenuItem(listener,"Bottom","view.tabsPosition.bottom"));

		submenu = menu.appendSubmenu("Zoom");
		submenu.append(new MenuItem(listener,"_In","view.zoom.in"));
		submenu.append(new MenuItem(listener,"_Out","view.zoom.out"));

		
		menu = menubar.append("_Tools");
		
		menu = menubar.append("_leds",true);
		menu.append(new MenuItem(listener,"_Contents","help.contents"));
		menu.append(new SeparatorMenuItem());
		menu.append(new MenuItem(listener,"_About","help.about"));
		
		return menubar;
	}

	/**
	 * Set the main leds toobar
	 * \todo report bug "forward references to class" on the D news group
	 */
	public Widget getToolbar(ButtonClickedListener listener)
	{
		HandleBox handleBox = new HandleBox();
		Toolbar toolbar = new Toolbar();
		//toolbar.setIconSize(IconSize.SMALL_TOOLBAR);
		toolbar.setIconSize(IconSize.MENU);
		toolbar.setStyle(ToolbarStyle.ICONS);
		Dispatcher dispatcher = Dispatcher.getDispatcher();
		
		// this exist because of compiler
		// problems on "forward references to classes"
		void addButton(Stock stockID, char[] tooltip, char[] action)
		{
			Button button = (Button)toolbar.insertStock(stockID,tooltip,"",-1);
			dispatcher.addButtonClickedListener(this,button,action);
		}
		
		addButton(Stock.NEW,"New","toolbar.new");
		addButton(Stock.OPEN,"Open","toolbar.open");
		addButton(Stock.REVERT_TO_SAVED,"Revert","toolbar.revert");
		addButton(Stock.CLOSE,"close","toolbar.close");
		toolbar.appendSpace();
		addButton(Stock.SAVE,"save","toolbar.save");
		addButton(Stock.SAVE_AS,"save as","toolbar.saveAs");
		toolbar.appendSpace();
		addButton(Stock.PRINT,"print","toolbar.print");
		toolbar.appendSpace();
		addButton(Stock.UNDO,"undo","toolbar.undo");
		addButton(Stock.REDO,"redo","toolbar.redo");
		toolbar.appendSpace();
		addButton(Stock.COPY,"copy","toolbar.copy");
		addButton(Stock.CUT,"cut","toolbar.cut");
		addButton(Stock.PASTE,"paste","toolbar.paste");
		toolbar.appendSpace();
		addButton(Stock.FIND,"find","toolbar.find");
		addButton(Stock.FIND_AND_REPLACE,"find and replace","toolbar.findAndReplace");
		addButton(Stock.GO_FORWARD,"goto line","toolbar.gotoLine");
		toolbar.appendSpace();
		addButton(Stock.EXECUTE,"executes the 'make' command for the selected project","toolbar.go");
		addButton(Stock.APPLY,"executes the 'Run command' for the selected project","toolbar.execute");
		addButton(Stock.STOP,"nothing for now (stop)","toolbar.stop");
		toolbar.appendSpace();
		Button button = new Button("find:");
		dispatcher.addButtonClickedListener(this,button,"toolbar.findText");
		
		toolbar.appendWidget(button,"Find text","");
		if ( findTextCombo === null )
		{
			findTextCombo = new Combo();
		}
		toolbar.appendWidget(findTextCombo,"text to find","");

		Menu menu = new Menu();
		menu.append(new MenuItem(this,"D","lexer.D"));
		menu.append(new MenuItem(this,"C","lexer.C"));
		menu.append(new MenuItem(this,"C++","lexer.CPP"));
		menu.append(new MenuItem(this,"html","lexer.html"));
		menu.append(new MenuItem(this,"Makefile","lexer.makefile"));
		menu.append(new MenuItem(this,"DUI","lexer.D"));
		OptionMenu optionMenu = new OptionMenu(menu);
		optionMenu.setMenu(menu);

		toolbar.appendWidget(optionMenu,"Language","");

		handleBox.add(toolbar);
		
		return handleBox;
		
	}

	/**
	 * Creates the standard views
	 */
	public Widget getViews()
	{
		
		dockLeft = new Docker("Utils");
		dockTop = new Docker("Top",true);
		
		dockRight = new Docker("Utils");
		dockBottom = new Docker("Consoles",true);
		dockCode = new Docker("Main view");
		dockCode.setScrollable(true);
		
		HPaned hOut = new HPaned();
		leftMainPane = new HPaned(dockLeft.getWidget(),dockCode.getWidget());
	
		hOut.add1(leftMainPane);//,dockRight);
		bottomMainPane = new VPaned();
		VPaned vIn = new VPaned;
		vIn.add2(hOut);
		bottomMainPane.add(vIn,dockBottom.getWidget());
		
		return bottomMainPane;
	}

	public ProjectView getProjectView()
	{
		if ( projectView === null )
		{
			projectView = new ProjectView(this);
		}
		return projectView;
	}
	
	public SystemConsole getSystemConsole()
	{
		if ( systemConsole === null )
		{
			systemConsole = new SystemConsole(this);
		}
		return systemConsole;
	}
	
	public void addConsole(SystemConsole console)
	{
		dockBottom.dock(console);
	}

	/**
	 * Creates the standard statusbar
	 */
	public StatusBar getStatusbar()
	{
		if ( statusbar === null )
		{
			statusbar = new StatusBar();
		}
		
		return statusbar;
	}

	/**
	 * Create the find window
	 */
	public FindAndReplace getFindAndReplace()
	{
		if ( findAndReplace === null )
		{
			findAndReplace = new FindAndReplace(this);
		}
		return findAndReplace;
	}

	/**
	 * Create the goto line dialog
	 */
	public GotoLineDialog getGotoLine()
	{
		if ( gotoLineDialog === null )
		{
			gotoLineDialog = new GotoLineDialog(this);
		}
		return gotoLineDialog;
	}
		
	public char[] getFindText()
	{
		if ( findTextCombo === null )
		{
			return null;
		}
		return findTextCombo.getSelectedText();
	}
	
	/**
	 * Menu listener callback
	 */
	void activateCallback(MenuItem menuItem, char [] action)
	{
		bit processed = false;
		// the most used action are upon the buffer so lets try that first
		FileView cv = getCurrentFileView();
		
		if ( cv !== null )
		{
			processed = cv.execAction(action);
		}
		
		if ( !processed )
		{
			switch(action)
			{

				case "file.newProject": new Project(this,""); break;
				case "file.open": openFiles() ; break;
				case "file.save": saveCurrent() ; break;
				case "file.saveAs": saveAsCurrent() ; break;
	
				case "file.exit":
					if ( exit(0,false) )
					{
						DUI.dui().stop();
					}
				break;

				case "edit.preferences":
					Leds.leds().editPreferences();
					break;
				
				case "view.toggleLeftPanel":
					dockLeft.toggleVisible();
					break;
	
				case "view.tabsPosition.left":	dockCode.setTabPos(PositionType.LEFT); break;
				case "view.tabsPosition.top":	dockCode.setTabPos(PositionType.TOP); break;
				case "view.tabsPosition.right":	dockCode.setTabPos(PositionType.RIGHT); break;
				case "view.tabsPosition.bottom":dockCode.setTabPos(PositionType.BOTTOM); break;
	
				case "help.about":
					MessageDialog d = new MessageDialog(
							this, 0,
							DialogType.INFO,DialogButtons.OK,
							"leds\n"
							"Linux Editor for D\n"
							"S as in simple (superfoolous) or supperb\n"
							" \n"
							"build with DUI\n"
							" \n"
							"by Antonio Monteiro\n"
							);
					d.run();
					d.destroy();
				break;
				
				default:
					MessageDialog d = new MessageDialog(this, 0,
					DialogType.INFO,DialogButtons.OK,"You pressed "~action);
					d.run();
					d.destroy();
				break;
			}
		}
	}
	
	void activateItemCallback(MenuItem menuItem, char [] action)
	{
		printf("Workspace.MenuListener.ctivateItemCallback %.*s\n",action);
	}

	/**
	 * Button listener callback.
	 * Used for the toolbar buttons
	 */
	void buttonClickedCallback(Button button, char [] action)
	{
		bit processed = false;
		// the most used action are upon the buffer so lets try that first
		FileView cv = getCurrentFileView();
		
		if ( cv !== null )
		{
			processed = cv.execAction(action);
		}
		
		if ( !processed )
		{
			Project project = projectView.getCurrentProject();
			switch(action)
			{
				case "toolbar.open": openFiles() ; break;
				case "toolbar.save": saveCurrent() ; break;
				case "toolbar.saveAs": saveAsCurrent() ; break;
				
				case "toolbar.go":
					projectView.runTool(project, "makeCommand:", true);
					break;

				case "toolbar.execute":
					projectView.runTool(project, "runCommand:", true);
					break;

				case "toolbar.stop":
					projectView.runTool(project, "stopCommand:", true);
					break;

				case "toolbar.new" :
					dockCode.dock(new CodeView(this));
					break;
	
				case "toolbar.close" :
					closeView();
					break;
	
				case "toolbar.find": // no break
				case "toolbar.findAndReplace":
					getFindAndReplace().show();
					break;
				
				default:
					MessageDialog d = new MessageDialog(this, 0,
					DialogType.INFO,DialogButtons.OK,"You pressed "~action);
					d.run();
					d.destroy();
				break;
			}
		}
	}

	FileView getCurrentFileView()
	{
		return dockCode.getCurrentFileView();
	}
	
	/**
	 * close the current view
	 */
	void closeView()
	{
		
		FileView fileView = getCurrentFileView();
		
		if ( fileView.allowClose() )
		{
			if ( !fileView.getEmpty() || dockCode.getNDocks()>1 )
			{
				dockCode.undockCurrent();
				if ( dockCode.getNDocks() == 0 )
				{
					dockCode.dock(new CodeView(this));
				}
			}
		}
	}

	/**
	 * closes the current view if it is empty
	 * and the number of pages if > 1.
	 * The main docker will always have a least one code buffer
	 */
	void closeIfEmpty()
	{
		FileView fileView = getCurrentFileView();
		if ( fileView.getEmpty() && dockCode.getNDocks()>1)
		{
			dockCode.undockCurrent();
		}
	}
	
	/**
	 * close all views
	 */
	void closeViewAll()
	{
	}

	/**
	 * creates a new Project
	 */
	void newProject()
	{
		
	}
	
	/**
	 * Adds a project to the current project view
	 * @param project the project to add
	 * \todo ??? add the project to the view that is interested on this type of project ???
	 */
	void addProject(Project project)
	{
		projectView.addProject(project);
	}
	
	/**
	 * Creates a new CodeVivew with the content of the file
	 */
	void openFile(char [] filename)
	{
		bit isImage = false;
		FileView fileView;
		
		if ( ImageView.isImageFile(filename) )
		{
			fileView = new ImageView(this,filename);
		}
		else
		{
			fileView = new CodeView(this,filename);
		}
		dockCode.dock(fileView);
		closeIfEmpty();
	}
	
	/**
	 * open files into the dockCode
	 * @param filesToOpen
	 */
	void openFiles(char[][] filesToOpen)
	{
		for ( int i=0 ; i<filesToOpen.length ; i++ )
		{
			openFile(filesToOpen[i]);
		}
	}
	
	/**
	 * open files into the dockCode
	 * \todo do we need to keep a ref in D to the FileViews?
	 */
	void openFiles()
	{
		if ( fileSelection === null )
		{
			fileSelection = new FileSelection("Open file");
		}
		fileSelection.complete("*.d");
		fileSelection.setMultiple(true);
		fileSelection.setTitle("Open file");

		if ( fileSelection.run() == Response.OK )
		{
			char [][] selections = fileSelection.getSelections();
			openFiles(selections);
			//for ( int i=0 ;i<selections.length ; i++)
			//{
			//	FileView fileView = new fileView(this,selections[i]);
			//	dockCode.dock(fileView,fileView);
			//}
			//closeIfEmpty();
		}
		fileSelection.hide();
	}

	/**
	 * Saves the current buffer
	 */
	void saveCurrent()
	{
		FileView fileView = getCurrentFileView();
		if ( fileView.getFileName().length == 0 )
		{
			saveAs(fileView,"Save file");
		}
		else
		{
			fileView.saveFile();
		}
				
	}
	
	/**
	 * Ask the user a new file name to save the current buffer.
	 */
	void saveAsCurrent()
	{
		FileView fileView = getCurrentFileView();
		saveAs(fileView,"Save file");
	}

	/**
	 * Ask the user a new file name to save the buffer
	 * @param fileView the file view to save
	 */
	void saveAs(FileView fileView, char[] title)
	{
		if ( fileSelection === null )
		{
			fileSelection = new FileSelection("Save file");
		}
		fileSelection.setMultiple(false);
		fileSelection.setTitle(title);

		fileSelection.complete(fileView.getBaseName());

		if ( fileSelection.run() == Response.OK )
		{
			char[] selection = fileSelection.getFileName();
			if ( fileView.saveFile(selection) )
			{
				dockCode.setDockedLabelText(fileView,fileView.getBaseName());
			}
		}
		fileSelection.hide();
	}	

	/**
	 * findReplace
	 */
	void findReplace(char[] findStr, char[] replaceStr, bit replace, bit find,int flags)
	{
		FileView cv = getCurrentFileView();
		if ( cv !== null )
		{
			cv.findReplace(findStr, replaceStr, replace, find,flags);
		}
	}

	void gotoLine(int lineNumber)
	{
		FileView cv = getCurrentFileView();
		if ( cv !== null )
		{
			cv.gotoLine(lineNumber);
		}
	}

	Leds getLeds()
	{
		return parentLeds;
	}
	
}
